import json

import pytest

from demoproject.app1.models import DemoModel1, DemoModel2
from djaxei.modems.model import FieldListModelMoDem


@pytest.mark.parametrize(
    'm1, m2',
    (
        pytest.param(
            'app1.demomodel1', 'app1.DemoModel2',
            id='strings'
        ),
        pytest.param(
            DemoModel1, DemoModel2,
            id='models'
        ),
        pytest.param(
            DemoModel1, 'app1.demomodel2',
            id='mixed'
        ),
    )
)
def test_source_models_flavour(m1, m2):
    exp1 = FieldListModelMoDem(
            model=m1,
            fields=['id']
        )
    exp2 = FieldListModelMoDem(
            model=m2,
            fields=['id']
        )
    assert exp1.field_list == ['id']
    assert exp2.field_list == ['id']
    assert exp1.model_label == 'app1.demomodel1'
    assert exp2.model_label == 'app1.demomodel2'


def test_field_lists_get_header():
    exp1 = FieldListModelMoDem(
        model=DemoModel1,
        fields=['id', 'any....', ('alfa', str), ('bravo', json.dumps)]
    )
    assert exp1.field_list == ['id', 'any....', ('alfa', str), ('bravo', json.dumps)]
    assert exp1.get_header() == ['id', 'any....', 'alfa', 'bravo']


def test_field_lists_modulate(demomodel1):
    fx_dt = lambda dt: dt.replace(tzinfo=None)

    flist = ['id', 'fk_id', 'char', 'integer', 'logic', 'null_logic',
                'date', 'nullable', 'choice',
                ('timestamp', fx_dt), ('j', json.dumps)]

    exp1 = FieldListModelMoDem(
        model=DemoModel1,
        fields=flist
    )
    assert exp1.field_list == flist
    assert exp1.modulate(demomodel1) == [
        demomodel1.id,
        demomodel1.fk_id,
        demomodel1.char,
        demomodel1.integer,
        demomodel1.logic,
        demomodel1.null_logic,
        demomodel1.date,
        demomodel1.nullable,
        demomodel1.choice,
        fx_dt(demomodel1.timestamp),
        json.dumps(demomodel1.j)
    ]


def test_field_list_mandatory(demomodel1):
    exp1 = FieldListModelMoDem(
        model=DemoModel1,
        fields=[]
    )
    with pytest.raises(RuntimeError, match='Field list is mandatory for modulate'):
        exp1.modulate(demomodel1)



import importlib
import random

import datetime

import factory
import logging

import pytest
from dateutil.tz import UTC
from factory.fuzzy import FuzzyInteger, FuzzyDate, FuzzyChoice, FuzzyText, FuzzyDateTime

from demoproject.app1.models import DemoModel1, DemoModel2, DemoModel3, VeryLongNameModelDemoModel4


@pytest.fixture(scope='session')
def client(request):
    import django_webtest
    wtm = django_webtest.WebTestMixin()
    wtm.csrf_checks = False
    wtm._patch_settings()
    request.addfinalizer(wtm._unpatch_settings)
    app = django_webtest.DjangoTestApp()
    return app


def pytest_configure():
    logger = logging.getLogger("djaxei")
    handler = logging.NullHandler()
    logger.handlers = [handler]


CHOICES_IDS = [x[0] for x in DemoModel1.CHOICES]
class DemoModel1Factory(factory.django.DjangoModelFactory):
    char = FuzzyText(length=10, prefix='dm1_')
    integer = FuzzyInteger(1000)
    date = FuzzyDate(datetime.date(2008, 1, 1))
    timestamp = FuzzyDateTime(start_dt=datetime.datetime(2008, 1, 1, 13, 44, 1, tzinfo=UTC))
    choice = FuzzyChoice(CHOICES_IDS)
    j = {
        'alfa': 1,
        'beta': {
            'gamma': 'delta'
        }
    }

    class Meta:
        model = DemoModel1


class DemoModel2Factory(factory.django.DjangoModelFactory):
    integer = FuzzyInteger(1000)
    char = FuzzyText(length=10, prefix='dm2_')

    class Meta:
        model = DemoModel2


class DemoModel3Factory(factory.django.DjangoModelFactory):
    integer = FuzzyInteger(1000)
    char = FuzzyText(length=10, prefix='dm3_')

    class Meta:
        model = DemoModel3


class VeryLongNameModelDemoModel4Factory(factory.django.DjangoModelFactory):
    integer = FuzzyInteger(1000)
    char = FuzzyText(length=10, prefix='dm4_')

    class Meta:
        model = VeryLongNameModelDemoModel4


class DemoModel5Factory(factory.django.DjangoModelFactory):
    integer = FuzzyInteger(1000)
    char = FuzzyText(length=10, prefix='dm5_')

    class Meta:
        model = VeryLongNameModelDemoModel4


@pytest.fixture
def demomodel1(db):
    ret = DemoModel1Factory.create()
    return ret


@pytest.fixture
def records1(db):
    ret = []
    for n in range(5):
        ret.append(DemoModel1Factory.create())
    return ret


@pytest.fixture
def records2(records1):
    ret = []
    for n in range(3):
        for master in records1:
            ret.append(DemoModel2Factory.create(fk=master))
    return ret


@pytest.fixture
def records3(records1):
    ret = []
    for n in range(3):
        for master in records1:
            ret.append(DemoModel3Factory.create(fk=master))
    return ret


@pytest.fixture
def records4(records2, records3):
    ret = []
    for n in range(15):
        master1 = random.choice(records2)
        master2 = random.choice(records3) if n % 3 != 0 else None
        ret.append(VeryLongNameModelDemoModel4Factory.create(fk2=master1, fk3=master2))
    return ret


@pytest.fixture
def records5(records2, records3):
    ret = []
    for n in range(15):
        master1 = random.choice(records2)
        master2 = random.choice(records3) if n % 3 != 0 else None
        ret.append(DemoModel5Factory.create(fk2=master1, fk3=master2))
    return ret


@pytest.fixture
def recordset(records4, records5):
    return records4 + records5



import json
from django.apps import apps
from openpyxl.worksheet.worksheet import Worksheet

from pathlib import Path

import pytest

from djaxei.imp import Importer
from demoproject.app1.models import DemoModel1, DemoModel2, DemoModel3, VeryLongNameModelDemoModel4
from djaxei.modems.field import DatetimeNonAwareModem, JsonToStringModem, RemapperFieldModem
from djaxei.modems.model import FieldListModelMoDem


def row_loader_1(storage):
    def _f(data_dict):  # zip(header, fields)
        id = data_dict.pop('id')
        try:
            obj = DemoModel1.objects.create(**data_dict)
        except Exception as e:
            print(e)
        storage.setdefault(1, {})[id] = obj.id

    return _f


def row_loader_2(storage):
    def _f(data_dict):  # zip(header, fields)
        id = data_dict.pop('id')
        data_dict['fk_id'] = storage[1][data_dict['fk_id']]
        obj = DemoModel2.objects.create(**data_dict)
        storage.setdefault(2, {})[id] = obj.id

    return _f


def row_loader_3(storage):
    def _f(data_dict):  # zip(header, fields)
        id = data_dict.pop('id')
        data_dict['fk_id'] = storage[1][data_dict['fk_id']]
        obj = DemoModel3.objects.create(**data_dict)
        storage.setdefault(3, {})[id] = obj.id

    return _f


def row_loader_4(storage):
    def _f(data_dict):  # zip(header, fields)
        id = data_dict.pop('id')
        # fk2_id may have not been remapped
        data_dict['fk2_id'] = storage[2].get(data_dict['fk2_id'], data_dict['fk2_id'])
        # fk3_id may have not been remapped
        if data_dict['fk3_id'] is not None:
            data_dict['fk3_id'] = storage[3].get(data_dict['fk3_id'], data_dict['fk3_id'])
        obj = VeryLongNameModelDemoModel4.objects.create(**data_dict)
        storage.setdefault(4, {})[id] = obj.id

    return _f


def loader_m1(mappings: dict, model_name: str, worksheet: Worksheet):
    klass = apps.get_model(*model_name.split('.'))

    for i, row in enumerate(worksheet.rows):
        if i == 0:  # header
            headers = [c.value for c in row][1:]
        else:
            oldid = row[0].value
            data = {h: row[z + 1].value for z, h in enumerate(headers)}

            data['j'] = json.loads(data['j'])

            new_obj = klass.objects.create(**data)
            mappings[model_name][oldid] = new_obj.id


def modifier_m1(row, *args):
    row['j'] = json.loads(row['j'])


def modifier_m2(row, mappings):
    row['fk_id'] = mappings['app1.demomodel2'][row['fk_id']]


def modifier_m3(row, mappings):
    row['fk_id'] = mappings['app1.demomodel3'][row['fk_id']]


def modifier_m4(row, mappings):
    row['fk_id'] = mappings['app1.verylongnamemodeldemomodel4'][row['fk_id']]


def loader_m2(mappings: dict, model_name: str, worksheet: Worksheet):
    klass = apps.get_model(*model_name.split('.'))

    for i, row in enumerate(worksheet.rows):
        if i == 0:  # header
            headers = [c.value for c in row][1:]
        else:
            oldid = row[0].value
            data = {h: row[z + 1].value for z, h in enumerate(headers)}

            data['fk_id'] = mappings['app1.demomodel1'][data['fk_id']]

            new_obj = klass.objects.create(**data)
            mappings[model_name][oldid] = new_obj.id


@pytest.mark.django_db
class TestImport(object):

    def test_importer(self, recordset):
        existing_records = {
            'app1.demomodel1': list(DemoModel1.objects.all()),
            'app1.demomodel2': list(DemoModel2.objects.all()),
            'app1.demomodel3': list(DemoModel3.objects.all()),
            'app1.verylongnamemodeldemomodel4': list(VeryLongNameModelDemoModel4.objects.all()),
        }

        modems = [
            FieldListModelMoDem(
                model='app1.demomodel1',
                fields=['id', RemapperFieldModem('fk_id', 'app1.demomodel1'), 'char', 'integer', 'logic', 'null_logic',
                        'date', 'nullable', 'choice',
                        DatetimeNonAwareModem('timestamp'), JsonToStringModem('j')]
            ),
            FieldListModelMoDem(
                model='app1.demomodel2',
                fields=['id', RemapperFieldModem('fk_id', 'app1.demomodel1'), 'integer']
            ),
            FieldListModelMoDem(
                model='app1.demomodel3',
                fields=['id', RemapperFieldModem('fk_id', 'app1.demomodel1'), 'char', 'integer']
            ),
            FieldListModelMoDem(
                model='app1.verylongnamemodeldemomodel4',
                fields=['id', RemapperFieldModem('fk3_id', 'app1.demomodel3'), 'char',
                        RemapperFieldModem('fk2_id', 'app1.demomodel2'), 'integer']
            ),
            # FieldListModelMoDem(
            #     'app1.demomodel2',
            #     ['id', 'fk_id', 'integer']
            # ),
            # FieldListModelMoDem(
            #     'app1.demomodel3',
            #     ['id', 'fk_id', 'char', 'integer']
            # ),
            # FieldListModelMoDem(
            #     'app1.verylongnamemodeldemomodel4',
            #     ['id', 'fk3_id', 'char', 'fk2_id', 'integer']
            # ),
        ]
        importer = Importer(modems=modems)

        ret = importer.xls_import(Path(__file__).parent / 'exported.xlsx')
        assert DemoModel1.objects.count() == len(existing_records['app1.demomodel1']) + 1
        assert DemoModel2.objects.count() == len(existing_records['app1.demomodel2']) + 4
        assert DemoModel3.objects.count() == len(existing_records['app1.demomodel3']) + 4
        assert VeryLongNameModelDemoModel4.objects.count() == len(existing_records['app1.verylongnamemodeldemomodel4']) + 11

    # def test_importer(self, records4):
    #     dm1 = list(DemoModel1.objects.all())
    #     dm2 = list(DemoModel2.objects.all())
    #     dm3 = list(DemoModel3.objects.all())
    #     dm4 = list(VeryLongNameModelDemoModel4.objects.all())
    #     fi = os.path.abspath(os.path.join(__file__, os.pardir, 'files', 'example_imp.xlsx'))
    #     storage = {}
    #     Importer().xls_import(fi,
    #       OrderedDict([
    #           ('demomodel1', row_loader_1(storage)),
    #           ('demomodel2', row_loader_2(storage)),
    #           ('demomodel3', row_loader_3(storage)),
    #           ('verylongnamemodeldemomodel4', row_loader_4(storage)),
    #       ])
    #     )
    #     assert len(dm1) == DemoModel1.objects.count() - 1
    #     assert len(dm2) == DemoModel2.objects.count() - 3
    #     assert len(dm3) == DemoModel3.objects.count() - 3
    #     assert len(dm4) == VeryLongNameModelDemoModel4.objects.count() - 7



# -*- coding: utf-8 -*-demomodel1
from __future__ import absolute_import, unicode_literals

from django.contrib.auth.models import User
from django.db import models


class DemoModel1(models.Model):
    CHOICES = ((1, 'Choice 1'), (2, 'Choice 2'), (3, 'Choice 3'))

    fk = models.ForeignKey(User,
                           on_delete=models.CASCADE,
                           blank=True, null=True)
    char = models.CharField(max_length=255)
    integer = models.IntegerField()
    logic = models.BooleanField(default=False)
    null_logic = models.NullBooleanField(default=None)
    date = models.DateField()
    timestamp = models.DateTimeField()
    nullable = models.CharField(max_length=255, null=True, default=None)
    choice = models.IntegerField(choices=CHOICES)
    j = models.JSONField(default={})



class DemoModel2(models.Model):
    fk = models.ForeignKey(DemoModel1,
                           on_delete=models.CASCADE,
                           blank=True, null=True)
    char = models.CharField(max_length=255, blank=True, null=True)
    integer = models.IntegerField(blank=True, null=True)


class DemoModel3(models.Model):
    fk = models.ForeignKey(DemoModel1,
                           on_delete=models.CASCADE)
    char = models.CharField(max_length=255, blank=True, null=True)
    integer = models.IntegerField(blank=True, null=True)


class VeryLongNameModelDemoModel4(models.Model):
    fk2 = models.ForeignKey(DemoModel2,
                           on_delete=models.CASCADE)
    fk3 = models.ForeignKey(DemoModel3,
                            blank=True, null=True,
                            on_delete=models.CASCADE)
    char = models.CharField(max_length=255, blank=True, null=True)
    integer = models.IntegerField(blank=True, null=True)


class DemoModel5(models.Model):
    fk2 = models.ForeignKey(DemoModel2,
                           on_delete=models.CASCADE)
    fk3 = models.ForeignKey(DemoModel3,
                            blank=True, null=True,
                            on_delete=models.CASCADE)
    char = models.CharField(max_length=255, blank=True, null=True)
    integer = models.IntegerField(blank=True, null=True)



from xlsxwriter import Workbook


class WorkbookImpl(Workbook):
    _preferred_suffix = ".xlsx"

    def __init__(self, filename=None, options=None):
        super(WorkbookImpl, self).__init__(filename, options)

    def write_data(self, data):
        for ws_name, rows in data.items():
            ws = self.add_worksheet(ws_name)
            for i, row in enumerate(rows):
                ws.write_row(i, 0, row)
        self.close()



"""The base ModelModems."""
from django.apps import apps
from django.db import models
from openpyxl.workbook import Workbook

from djaxei.modems.field import BaseFieldMoDem

TABNAME_LIMIT = 31


class AbstractModelMoDem:
    def __init__(self, model, *args, **kwargs):
        """Model is either a Django Model or the model._meta.label_lower of a Django model."""
        if hasattr(model, '_meta'):
            self.model_label = model._meta.label_lower
        else:  # Must be the label_lower
            self.model_label = model.lower()
        self._extra_args = {
            'args': args,
            'kwargs': kwargs
        }

    def modulate(self, obj: models.Model, context):
        pass

    def demodulate(self, obj, context):
        pass


class FieldListModelMoDemImporter:

    def __init__(self, remappings: dict, wb: Workbook, modem: AbstractModelMoDem) -> None:
        self.remappings = remappings
        self.worksheet = wb[modem.model_label[:TABNAME_LIMIT]]
        self.modem = modem
        self.manager = apps.get_model(modem.model_label).objects

    def run(self):
        pkpos, oldpk = None, None
        for i, row in enumerate(self.worksheet.rows):
            if i == 0:  # header
                headers = [c.value for c in row]
                if self.modem.pk in headers:
                    pkpos = headers.index(self.modem.pk)
                    headers.pop(pkpos)
                    self.modem.field_list.pop(pkpos)
            else:
                row = list(row)
                if pkpos is not None:
                    oldpk = row.pop(pkpos).value

                data = {}
                for z, h in enumerate(headers):
                    field = self.modem.field_list[z]
                    data[h] = row[z].value
                    if isinstance(field, BaseFieldMoDem):
                        data[h] = field.demodulate(data[h], self.remappings)

                new_obj = self.manager.create(**data)
                self.remappings[self.modem.model_label[:TABNAME_LIMIT]][oldpk] = new_obj.id


class FieldListModelMoDem(AbstractModelMoDem):
    """A ModelModem requiring a list of fields to serialize.

    The field in the list can be either be the str name of the field or a tuple consisting in:
    - the str name of the field
    - a function that will be passed the obj and the field name
    """
    def __init__(self, model, fields: list = None, pk='id', *args, **kwargs):
        super().__init__(model, *args, **kwargs)
        self.field_list = fields
        self.pk = pk
        self._importer = None

    def get_importer(self, remappings: dict, wb: Workbook):
        if self._importer is None:
            self._importer = FieldListModelMoDemImporter(remappings, wb, self)
        return self._importer

    def get_header(self):
        ret = []
        for fname in self.field_list:
            if isinstance(fname, str):
                val = fname
            elif hasattr(fname, 'get_field_header'):
                val = fname.get_field_header()
            elif isinstance(fname, (tuple, list)):
                val = fname[0]
            ret.append(val)
        return ret

    def modulate(self, obj):
        """Serialise obj using the provided field_list in Modem init.

        Raise exception if no field_list provided.
        """
        if not self.field_list:
            raise RuntimeError('Field list is mandatory for modulate')
        row = []
        for field in self.field_list:
            if isinstance(field, str):
                row.append(getattr(obj, field))
            elif hasattr(field, 'modulate'):
                row.append(field.modulate(obj))
            elif isinstance(field, (tuple, list)):
                row.append(field[1](getattr(obj, field[0])))
        return row

    def demodulate(self, obj, context):
        super().demodulate(obj, context)




import importlib


def get_workbook_impl():
    errors = []
    try:
        provider = importlib.import_module('djaxei.providers.xlsxwriter_provider')
        return provider.WorkbookImpl
    except Exception as e1:
        errors.append(e1)

    try:
        provider = importlib.import_module('djaxei.providers.xlwt_provider')
        return provider.WorkbookImpl
    except Exception as e2:
        errors.append(e2)

    errors = "\n".join(map(str, errors))
    raise RuntimeError(f'No provider found\n{errors}')



import datetime

import pytest
import pytz

from djaxei.modems.field import DatetimeNonAwareModem


class Dummy:
    pass

timezone = pytz.timezone("America/Los_Angeles")

@pytest.mark.parametrize(
    'input, expected',
    (
            pytest.param(
                datetime.datetime(year=2022, month=7, day=11, hour=14, minute=15, second=26,
                                  microsecond=324430, tzinfo=timezone), datetime.datetime(2022, 7, 11, 14, 15, 26),
                id='simple'
            ),
            pytest.param(
                datetime.datetime(year=2022, month=7, day=11, hour=14, minute=15, second=26,
                                  microsecond=324430), datetime.datetime(2022, 7, 11, 14, 15, 26),
                id='non-TZ-aware'
            ),
            pytest.param(
                datetime.datetime(year=2022, month=7, day=11, hour=14, minute=15, second=26, tzinfo=timezone), datetime.datetime(2022, 7, 11, 14, 15, 26),
                id='no-microseconds'
            )
    )
)
def test_valid_not_aware_datetime_modulate(input, expected):
    obj = Dummy()

    obj.example_field = input
    output = DatetimeNonAwareModem('example_field').modulate(obj)
    assert output == expected

    # NB: there is loss of information as DatetimeNonAwareModem does not support microseconds and TZ
    assert DatetimeNonAwareModem('example_field').demodulate(output) == input.replace(tzinfo=None, microsecond=0)



# Generated by Django 3.2.17 on 2023-02-02 21:33

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='DemoModel1',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('char', models.CharField(max_length=255)),
                ('integer', models.IntegerField()),
                ('logic', models.BooleanField(default=False)),
                ('null_logic', models.NullBooleanField(default=None)),
                ('date', models.DateField()),
                ('timestamp', models.DateTimeField()),
                ('nullable', models.CharField(default=None, max_length=255, null=True)),
                ('choice', models.IntegerField(choices=[(1, 'Choice 1'), (2, 'Choice 2'), (3, 'Choice 3')])),
                ('j', models.JSONField(default={})),
                ('fk', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='DemoModel2',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('char', models.CharField(blank=True, max_length=255, null=True)),
                ('integer', models.IntegerField(blank=True, null=True)),
                ('fk', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='app1.demomodel1')),
            ],
        ),
        migrations.CreateModel(
            name='DemoModel3',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('char', models.CharField(blank=True, max_length=255, null=True)),
                ('integer', models.IntegerField(blank=True, null=True)),
                ('fk', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='app1.demomodel1')),
            ],
        ),
        migrations.CreateModel(
            name='DemoModel5',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('char', models.CharField(blank=True, max_length=255, null=True)),
                ('integer', models.IntegerField(blank=True, null=True)),
                ('fk2', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='app1.demomodel2')),
                ('fk3', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='app1.demomodel3')),
            ],
        ),
        migrations.CreateModel(
            name='VeryLongNameModelDemoModel4',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('char', models.CharField(blank=True, max_length=255, null=True)),
                ('integer', models.IntegerField(blank=True, null=True)),
                ('fk2', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='app1.demomodel2')),
                ('fk3', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='app1.demomodel3')),
            ],
        ),
    ]



import pytest

from djaxei.modems.field import JsonToStringModem

class Dummy:
    pass


@pytest.mark.parametrize(
    'input, expected',
    (
            pytest.param(
                {'A': 1}, '{"A": 1}',
                id='simple'
            ),
            pytest.param(
                {}, '{}',
                id='empty'
            ),
            pytest.param(
                'abc123', '"abc123"',
                id='string'
            ),
    )
)
def test_valid_json_modulate(input, expected):
    obj = Dummy()

    obj.example_field = input
    output = JsonToStringModem('example_field').modulate(obj)
    assert output == expected

    assert JsonToStringModem('example_field').demodulate(output) == input



"""The base FieldModems."""
from datetime import datetime

import json

from django.db import models


class BaseFieldMoDem:
    def __init__(self, field_ref: str) -> None:
        self.field_ref = field_ref

    def get_field_header(self):
        return self.field_ref

    def modulate(self, obj: models.Model, mappings=None):
        """Modulate the value of the self.field_ref field of the provided Django model object."""
        return getattr(obj, self.field_ref)

    def demodulate(self, value, mappings=None):
        """Demodulate the self.field_ref value."""
        return value


class RemapperFieldModem(BaseFieldMoDem):

    def __init__(self, field_ref: str, mapping_key: str) -> None:
        super().__init__(field_ref)
        self.mapping_key = mapping_key

    def demodulate(self, value, mappings):
        ret = value
        if ret is not None:
            ret = mappings[self.mapping_key][value]
        return ret


class JsonToStringModem(BaseFieldMoDem):
    def modulate(self, obj: models.Model, mappings=None):
        return json.dumps(getattr(obj, self.field_ref))

    def demodulate(self, value, mappings=None):
        return json.loads(value)


class DatetimeNonAwareModem(BaseFieldMoDem):
    """A modem for nomalizin a TZ-aware datetime to a non-TZ-aware datetime.

    NB: there is loss of information as does not support microseconds and TZ
    """
    def modulate(self, obj: models.Model, mappings=None):
        dt: datetime = getattr(obj, self.field_ref)
        return dt.replace(tzinfo=None, microsecond=0)



from xlwt import Workbook


class WorkbookImpl(Workbook):
    _preferred_suffix = ".xls"

    def __init__(self, filename, **kwargs):
        encoding, style_compression = kwargs.get('encoding', None), kwargs.get('style_compression', None)
        super(WorkbookImpl, self).__init__(encoding, style_compression)
        self._filename = filename

    def write_data(self, data):
        for ws_name, rows in data.items():
            ws = self.add_sheet(ws_name)
            for i, row in enumerate(rows):
                for j, value in enumerate(row):
                    if value is not None:
                        ws.write(i, j, unicode(value))
        self.save(self._filename)



class ImportException(Exception):
    def __init__(self, cause, worksheet=None, *args):
        super(ImportException, self).__init__(*args)
        self.worksheet = worksheet
        self.cause = cause

    def __str__(self):
        if hasattr(self.cause, 'reference'):
            msg = "[%s:%s] %s: %s" % (self.worksheet, self.cause.reference, self.cause.__class__.__name__, str(self.cause))
        else:
            msg = "[%s] %s: %s" % (self.worksheet, self.cause.__class__.__name__, str(self.cause))
        return msg




import os
from collections import OrderedDict
from collections.abc import Iterable
from tempfile import NamedTemporaryFile

from django.contrib.admin.utils import NestedObjects
from django.db import router, models
from django.db.models.query import QuerySet
from openpyxl import Workbook


class AbstractMoDem:

    def __init__(self, model, rules) -> None:
        """The base exporter class.

        :param model:
        :param rules:
        """
        self.rules = rules
        if isinstance(model, str):  # we assume 'app.modelname' format
            self.model = model.strip().lower()
        elif isinstance(model, models.Model):
            self.model = model._meta.label_lower

    def modulate(self, obj: models.Model, context):
        pass

    def demodulate(self, obj, context):
        pass


class SimpleMoDem:
    """A simple MoDem class accepting e list of either directly fieldnames or (fieldname, FieldMoDem).

    :param model:
    :param rules:
    """
    def modulate(self, obj: models.Model, context):
        row = []
        for x in self.rules:
            if isinstance(x, str):
                row.append(getattr(obj, x))
            else:  # we consider FieldMoDem
                row.append(x[1]().modulate(getattr(obj, x[0])))
        return row



class ExcelMoDem(AbstractMoDem):

    def modulate(self, obj: models.Model, context):
        fields = obj._meta.label_lower
        row = []
        for x in fields:
            if isinstance(x, str):
                row.append(getattr(obj, x))
            else:
                row.append(x[1](getattr(obj, x[0])))
        return row


class Exporter:
    def __init__(self, root, modems: list):
        """Create an exporter.

        :param root: Can be a Django object, or a queryset
        :param rules: A dictionary 'appname.modelname': list of field names or tuple(fieldname, stringify function)
        """
        if isinstance(root, QuerySet):
            self.roots = root
            self.using = router.db_for_write(root.first()._meta.model)
        elif isinstance(root, Iterable):
            self.roots = root
            self.using = router.db_for_write(list(root)[0]._meta.model)
        else:
            self.roots = [root]
            self.using = router.db_for_write(root._meta.model)
        self.modems = modems

        # {}
        # for model_ref, field_refs in rules.items():
        #     if isinstance(model_ref, str):
        #         model_ref = model_ref.lower()
        #     else:
        #         model_ref = model_ref._meta.label_lower
        #     self.mode[model_ref] = field_refs

    def get_modem_for_object(self, obj):
        ret = None
        for modem in self.modems:
            if modem.model_label == obj._meta.label_lower:
                ret = modem
                break
        return ret


    def xls_export(self, target):
        # lmodels = {}

        wb = Workbook(write_only=True)
        sheets = OrderedDict()

        for modem in self.modems:
            header = modem.get_header()
            sheets[modem.model_label] = [header, ]
            wb.create_sheet(modem.model_label)

        collector = NestedObjects(using=self.using)
        collector.collect(self.roots)

        def callback_generator(exporter):
            exporter = exporter
            def callback(obj):
                modem = exporter.get_modem_for_object(obj)
                sheets[obj._meta.label_lower].append(modem.modulate(obj))

                # fields = lmodels.get(obj._meta.label_lower, None)
                # if fields:
                #     row = []
                #     for x in fields:
                #         if isinstance(x, str):
                #             row.append(getattr(obj, x))
                #         else:
                #             row.append(x[1](getattr(obj, x[0])))
                #     sheets[obj._meta.label_lower].append(row)
            return callback

        collector.nested(callback_generator(self))

        for sheet_name, rows in sheets.items():
            for row in rows:
                wb[sheet_name].append(row)

        wb.save(target)
        #
        # workbook = None
        # workbookfile = None
        # try:
        #     sheets = OrderedDict()
        #
        #     lmodels = {}
        #     for k, v in _models.items():
        #         lname = k.lower()
        #         model_name = lname.rsplit('.')[1]
        #         lmodels[lname] = v
        #         sheets[model_name] = [v, ]
        #
        #     if root:
        #         root_qs = root._meta.model.objects.filter(pk=root.pk)
        #
        #     using = router.db_for_write(root_qs.first()._meta.model)
        #     collector = NestedObjects(using=using)
        #     collector.collect(root_qs)
        #
        #     def callback(obj):
        #         fields = lmodels.get(obj._meta.label_lower, None)
        #         if fields:
        #             sheets[obj._meta.model_name].append([getattr(obj, x) for x in fields])
        #
        #     collector.nested(callback)
        #
        #     Workbook = get_workbook_impl()
        #     workbookfile = self.dest or NamedTemporaryFile(dir=self.tmpdir, suffix=Workbook._preferred_suffix, delete=False)
        #     Workbook(workbookfile).write_data(sheets)
        #
        #     return workbookfile.name
        #
        # except Exception as e:
        #     if workbook:
        #         if not workbookfile.closed:
        #             workbookfile.close()
        #         if os.path.exists(workbookfile.name):
        #             os.remove(workbookfile.name)
        #     raise e


from openpyxl.reader.excel import load_workbook


class Importer:
    def __init__(self, modems: list):
        self.modems = modems

    def xls_import(self, file):
        wb = load_workbook(file, read_only=True, data_only=True)

        remappings = {
            sheet_name: {} for sheet_name in wb.sheetnames
        }

        for modem in self.modems:
            modem.get_importer(remappings, wb).run()

#
# class Importer_:
#     def __init__(self, tmpdir=None, **kwargs):
#         self.tmpdir = tmpdir
#
#     def xls_import(self, file, models_dict, *args, **kwargs):
#         wb = load_workbook(file, data_only=True)
#
#         for ws_name, callback in models_dict.items():
#             ws = wb[ws_name]
#             for rownum, row in enumerate(ws.iter_rows()):
#                 row = [cell.value for cell in row]
#                 if rownum == 0:
#                     header = row
#                 else:
#                     try:
#                         data_dict = OrderedDict(zip(header, row))
#                         data_dict = {k: v for k, v in data_dict.items() if k != None}
#                         callback(data_dict)
#                     except Exception as e:
#                         raise ImportException(e, worksheet=ws_name)
#         wb.close()




# see https://github.com/pytest-dev/pytest-django/blob/master/pytest_django/fixtures.py
import datetime
import json
import random
from collections.abc import Iterable
from tempfile import NamedTemporaryFile

import pytest
from demoproject.app1.models import (DemoModel1, DemoModel2,
                                     DemoModel3, VeryLongNameModelDemoModel4,)
from django.db.models import Q, QuerySet
from openpyxl.reader.excel import load_workbook

from djaxei.exp import Exporter
from djaxei.modems.field import DatetimeNonAwareModem, JsonToStringModem
from djaxei.modems.model import FieldListModelMoDem


def root_generator(key):
    qs = DemoModel1.objects.all()
    return {
        1: random.choice(qs),
        2: qs.filter(id__in=random.sample(list(qs.values_list('id', flat=True)), 2)),
        3: random.sample(list(qs), 2)
    }[key]



@pytest.mark.parametrize(
    'root_fx_key, m1, m2, m3, m4',
    (
        pytest.param(
            1, 'app1.demomodel1', 'app1.demomodel2', 'app1.DemoModel3', 'app1.VeryLongNameModelDemoModel4',
            id='strings'
        ),
        pytest.param(
            2, DemoModel1, DemoModel2, DemoModel3, VeryLongNameModelDemoModel4,
            id='models'
        ),
        pytest.param(
            3, DemoModel1, 'app1.demomodel2', DemoModel3, VeryLongNameModelDemoModel4,
            id='mixed'
        ),
    )
)
def test_exporter(root_fx_key, m1, m2, m3, m4, recordset):
    roots = root_generator(root_fx_key)

    fx_dt = lambda dt: dt.replace(tzinfo=None)
    modems = {
        FieldListModelMoDem(
            model=m1,
            fields=['id', 'fk_id', 'char', 'integer', 'logic', 'null_logic',
                    'date', 'nullable', 'choice',
                    DatetimeNonAwareModem('timestamp'), JsonToStringModem('j')]
        ),
        FieldListModelMoDem(
            m2,
            ['id', 'fk_id', 'integer']
        ),
        FieldListModelMoDem(
            m3,
            ['id', 'fk_id', 'char', 'integer']
        ),
        FieldListModelMoDem(
            m4,
            ['id', 'fk3_id', 'char', 'fk2_id', 'integer']
        ),
    }

    exporter = Exporter(root=roots, modems=modems)
    with NamedTemporaryFile(suffix='.xlsx') as fo:
        exporter.xls_export(fo)

        # Checks starts here

        # Read generated workbook
        results = {}
        wb = load_workbook(filename=fo.name, read_only=True)
        assert sorted(wb.sheetnames) == ['app1.demomodel1', 'app1.demomodel2', 'app1.demomodel3', 'app1.verylongnamemodeldemomodel4']

        for sheet in wb.worksheets:
            results[sheet.title] = []
            for row in sheet.rows:
                results[sheet.title].append([c.value for c in row])

    if isinstance(roots, QuerySet):
        roots = list(roots.all())
    elif not isinstance(roots, Iterable):
        roots = [roots]
    root_class = roots[0].__class__
    root_model_name = root_class._meta.label_lower

    result_root_header = results[root_model_name][0]

    # check roots
    expected = list(root_class.objects.filter(id__in=[c.id for c in roots]).values())
    for e in expected:
        e['j'] = json.dumps(e['j'])
        e['timestamp'] = fx_dt(e['timestamp']).replace(microsecond=0)
        e['date'] = datetime.datetime.fromordinal(e['date'].toordinal())
    actual = [dict(zip(result_root_header, row)) for row in results[root_model_name][1:]]
    for k in actual:
        k['timestamp'] = k['timestamp'].replace(microsecond=0)
    assert sorted(expected, key=lambda x: x['id']) == sorted(actual, key=lambda x: x['id'])

    root_ids = [x.id for x in roots]

    # check DemoModel2
    ids2 = sorted([x[0] for x in results['app1.demomodel2'][1:]])
    assert sorted(list(DemoModel2.objects.filter(fk_id__in=root_ids).values_list('id', flat=True))) == ids2

    # check DemoModel3
    ids3 = sorted([x[0] for x in results['app1.demomodel3'][1:]])
    assert sorted(list(DemoModel3.objects.filter(fk_id__in=root_ids).values_list('id', flat=True))) == ids3

    # check VeryLongNameModelDemoModel4
    ids4 = sorted([x[0] for x in results['app1.verylongnamemodeldemomodel4'][1:]])
    assert sorted(
        list(VeryLongNameModelDemoModel4.objects.filter(Q(fk3__fk__in=root_ids) | Q(fk2__fk__in=root_ids)).values_list('id', flat=True))
    ) == ids4



import pytest
import random

from demoproject.app1.models import DemoModel1, VeryLongNameModelDemoModel4, DemoModel2, DemoModel3
from djaxei.exp import Exporter


def root_generator(key):
    qs = DemoModel1.objects.all()
    return {
        1: random.choice(qs),
        2: qs.filter(id__in=random.sample(list(qs.values_list('id', flat=True)), 2)),
        3: random.sample(list(qs), 2)
    }[key]



@pytest.mark.parametrize(
    'root_fx_key',
    (
        pytest.param(
            1,
            id='strings'
        ),
        pytest.param(
            2,
            id='models'
        ),
        pytest.param(
            3,
            id='mixed'
        ),
    )
)
def test_exporter_normalize_roots_to_list_of_objects(root_fx_key, recordset):
    roots = root_generator(root_fx_key)
    exporter = Exporter(root=roots, modems=[])

    if root_fx_key == 1:
        assert exporter.roots == [roots]
    elif root_fx_key == 2:
        assert exporter.roots == roots
    elif root_fx_key == 3:
        assert exporter.roots == roots
